#!/usr/bin/env node
"use strict";
var __create = Object.create;
var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __getProtoOf = Object.getPrototypeOf;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __copyProps = (to, from, except, desc) => {
  if (from && typeof from === "object" || typeof from === "function") {
    for (let key of __getOwnPropNames(from))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from[key], enumerable: !(desc = __getOwnPropDesc(from, key)) || desc.enumerable });
  }
  return to;
};
var __toESM = (mod, isNodeMode, target) => (target = mod != null ? __create(__getProtoOf(mod)) : {}, __copyProps(
  // If the importer is in node compatibility mode or this is not an ESM
  // file that has been converted to a CommonJS file using a Babel-
  // compatible transform (i.e. "__esModule" has not been set), then set
  // "default" to the CommonJS "module.exports" for node compatibility.
  isNodeMode || !mod || !mod.__esModule ? __defProp(target, "default", { value: mod, enumerable: true }) : target,
  mod
));

// src/index.ts
var import_websocket_polyfill = require("websocket-polyfill");

// src/main.ts
var import_fs_extra3 = require("fs-extra");

// src/console.ts
var import_colorette = require("colorette");
var info = (message = "") => console.log((0, import_colorette.dim)(message));
var warn = (message = "") => console.warn((0, import_colorette.yellow)(message));
var success = (message = "") => console.log((0, import_colorette.green)(message));
var error = (message = "") => console.error((0, import_colorette.bold)((0, import_colorette.red)(message)));
var die = (errorOrMessage, instructions) => {
  if (errorOrMessage instanceof Error) {
    error(errorOrMessage.message);
    info(errorOrMessage.stack);
  } else {
    error(errorOrMessage);
    if (instructions) {
      info(instructions);
    }
  }
};

// src/utils.ts
var getEmojiForPlatform = (platform) => {
  switch (platform) {
    case "ios":
      return "\u{1F34E}";
    case "android":
      return "\u{1F916}";
    default:
      return "\u{1F4F1}";
  }
};

// src/config.ts
var TOOL_CONFIG_DIR = "./.storybook-visual";
var STORYBOOK_CONFIG_PATH = "./.storybook";
var SUPPORTED_PLATFORMS = ["ios", "android"];
var SCREENSHOT_FORMAT = "png";
var SCREENSHOTS_CURRENT_DIR = ".current";
var SCREENSHOTS_BASE_DIR = ".base";
var SCREENSHOTS_DIFF_DIR = ".diff";
var SCREENSHOTS_DIRS = {
  current: `${TOOL_CONFIG_DIR}/${SCREENSHOTS_CURRENT_DIR}`,
  base: `${TOOL_CONFIG_DIR}/${SCREENSHOTS_BASE_DIR}`,
  diff: `${TOOL_CONFIG_DIR}/${SCREENSHOTS_DIFF_DIR}`
};
var serverConfig = {
  secured: false,
  host: "localhost",
  port: 7007,
  absolute: true
};
var serverVariables = {
  websocketType: serverConfig.secured ? "wss" : "ws",
  domain: `${serverConfig.host}:${serverConfig.port}`
};
var SERVER_DEFAULTS = {
  ...serverConfig,
  ...serverVariables,
  url: `${serverVariables.websocketType}://${serverVariables.domain}`
};

// src/cli.ts
var import_yargs = __toESM(require("yargs"));
var import_helpers = require("yargs/helpers");
var import_figlet = __toESM(require("figlet"));
var import_colorette2 = require("colorette");
var sayHi = async () => new Promise((resolve, reject) => {
  (0, import_figlet.default)("Storybook\nVisual Testing", (error2, data) => {
    if (error2) {
      reject(error2);
    }
    console.log((0, import_colorette2.red)(data ?? ""));
    resolve(data);
  });
});
var COMMANDS = {
  INIT: "init",
  TEST: "test",
  UPDATE: "update"
};
var getCliOptions = async () => {
  const argv = await (0, import_yargs.default)((0, import_helpers.hideBin)(process.argv)).version(false).strict(true).scriptName("react-native-storybook-visual").usage("Usage: $0 <command> [options]").command({
    command: COMMANDS.INIT,
    describe: "Initialize visual regression tool",
    builder: (yargs2) => yargs2.usage(`Usage: $0 ${COMMANDS.INIT} [options]`).example(`$0 ${COMMANDS.INIT}`, "Initialize visual regression tool").option("platform", {
      type: "array",
      default: SUPPORTED_PLATFORMS,
      description: "Platforms to test.",
      choices: SUPPORTED_PLATFORMS
    }).option("force", {
      type: "boolean",
      default: false,
      description: "Force re-initialization"
    }),
    handler: () => void 0
  }).command({
    command: COMMANDS.TEST,
    describe: "Test visual regression stories",
    builder: (yargs2) => yargs2.usage(`Usage: $0 ${COMMANDS.TEST} [options]`).example(`$0 ${COMMANDS.TEST}`, "Test visual regression stories").option("platform", {
      type: "array",
      default: SUPPORTED_PLATFORMS,
      description: "Platforms to test.",
      choices: SUPPORTED_PLATFORMS
    }).option("exitOnError", {
      type: "boolean",
      default: false,
      description: "Exits on first fail"
    }).option("files", {
      type: "array",
      description: "Specific files to test",
      defaultDescription: "`stories` field in your storybook config"
    }),
    handler: () => void 0
  }).command({
    command: COMMANDS.UPDATE,
    describe: "Update reference visual test screenshots",
    builder: (yargs2) => yargs2.usage(`Usage: $0 ${COMMANDS.UPDATE} [options]`).example(
      `$0 ${COMMANDS.UPDATE}`,
      "Update reference visual test screenshots"
    ).option("platform", {
      type: "array",
      default: SUPPORTED_PLATFORMS,
      description: "Platforms to test.",
      choices: SUPPORTED_PLATFORMS
    }),
    handler: () => void 0
  }).demandCommand(1, 1, "Please specify a valid command").parseAsync();
  const platforms = argv.platform;
  if (platforms.length === 1) {
    warn(
      `Running only on ${platforms[0]} ${getEmojiForPlatform(platforms[0])}`
    );
  }
  const options = {
    platforms,
    files: "files" in argv && Array.isArray(argv.files) ? argv.files : void 0,
    exitOnError: "exitOnError" in argv && typeof argv.exitOnError === "boolean" ? argv.exitOnError : void 0,
    watch: "watch" in argv && typeof argv.watch === "boolean" ? argv.watch : void 0,
    force: "force" in argv && typeof argv.force === "boolean" ? argv.force : void 0
  };
  return {
    command: argv._[0],
    options
  };
};

// src/devices.ts
var reverseAdbPort = async (port) => {
  try {
    const { execaCommand } = await import("execa");
    const { stderr } = await execaCommand(
      `adb reverse tcp:${port} tcp:${port}`
    );
    if (stderr) {
      die("\u{1F534} ADB reverse failed", stderr);
      process.exit(1);
    }
    return;
  } catch (error2) {
    die(
      "\u{1F534} ADB reverse failed",
      "message" in error2 ? error2.message : JSON.stringify(error2)
    );
    process.exit(1);
  }
};
var shutdownAndroidEmulator = async () => {
  try {
    const { execaCommand } = await import("execa");
    const { stderr } = await execaCommand(`adb emu kill`);
    if (stderr) {
      die("\u{1F534} ADB emulator kill failed", stderr);
    }
    return;
  } catch (error2) {
    die(
      "\u{1F534} ADB emulator kill failed",
      "message" in error2 ? error2.message : JSON.stringify(error2)
    );
  }
};
var shutdownBootedIosDevices = async () => {
  try {
    const { execaCommand } = await import("execa");
    const { stderr } = await execaCommand(`xcrun simctl shutdown all`);
    if (stderr) {
      die("\u{1F534} iOS simulator shutdown failed", stderr);
    }
    return;
  } catch (error2) {
    die(
      "\u{1F534} iOS simulator shutdown failed",
      "message" in error2 ? error2.message : JSON.stringify(error2)
    );
  }
};

// src/fs.ts
var import_promises = require("fs/promises");
var import_fs_extra = require("fs-extra");
var ensureRelativePathHasDot = (relativePath) => relativePath.startsWith(".") ? relativePath : `./${relativePath}`;
var getScreenshotDir = (type) => SCREENSHOTS_DIRS?.[type] ?? void 0;
var getScreenshotPath = ({
  type,
  platform,
  name
}) => {
  const dir = getScreenshotDir(type);
  return `${dir}/${platform}-${name}.${SCREENSHOT_FORMAT}`;
};
var isDirEmpty = async (path2) => {
  try {
    const files = await (0, import_promises.readdir)(path2);
    return files.length === 0;
  } catch (err) {
    return false;
  }
};
var isFileExists = async (path2) => {
  try {
    const exists = await (0, import_fs_extra.pathExists)(path2);
    return exists;
  } catch (err) {
    return false;
  }
};
var createDir = async (path2) => {
  try {
    await (0, import_fs_extra.ensureDir)(path2);
    return true;
  } catch (error2) {
    console.error("Could't create directory: ", path2, error2);
    return false;
  }
};
var createFile = async (path2, content) => {
  try {
    await (0, import_fs_extra.outputFile)(path2, content, { encoding: "utf-8" });
    return true;
  } catch (error2) {
    console.error("Could't create file: ", path2, error2);
    return false;
  }
};

// src/commands/update.ts
var import_promises2 = require("fs/promises");
var import_fs_extra2 = require("fs-extra");
var filterScreenshotFile = (file, platform) => file.startsWith(`${platform}-`) && file.endsWith(`.${SCREENSHOT_FORMAT}`);
var update = async (platforms) => {
  for (const platform of platforms) {
    info(
      `\u{1F501}${getEmojiForPlatform(
        platform
      )} Updating base screenshots for ${platform}`
    );
    try {
      const files = await (0, import_promises2.readdir)(SCREENSHOTS_DIRS.base);
      const filteredFiles = files.filter(
        (file) => filterScreenshotFile(file, platform)
      );
      for (const file of filteredFiles) {
        await (0, import_fs_extra2.remove)(`${SCREENSHOTS_DIRS.base}/${file}`);
      }
    } catch (error2) {
      die(
        `\u{1F534}${getEmojiForPlatform(
          platform
        )} Failed to remove old base screenshots for ${platform}`,
        "message" in error2 ? error2.message : JSON.stringify(error2)
      );
      process.exit(1);
    }
    try {
      const files = await (0, import_promises2.readdir)(SCREENSHOTS_DIRS.current);
      const filteredFiles = files.filter(
        (file) => filterScreenshotFile(file, platform)
      );
      for (const file of filteredFiles) {
        await (0, import_fs_extra2.move)(
          `${SCREENSHOTS_DIRS.current}/${file}`,
          `${SCREENSHOTS_DIRS.base}/${file}`,
          { overwrite: true }
        );
      }
    } catch (error2) {
      die(
        `\u{1F534}${getEmojiForPlatform(
          platform
        )} Failed to move base screenshots for ${platform}`,
        "message" in error2 ? error2.message : JSON.stringify(error2)
      );
      process.exit(1);
    }
  }
  success(`\u{1F7E2} Base screenshots updated`);
  process.exit(0);
};

// src/runner.ts
var import_node_events = require("events");
var import_addons = require("@storybook/addons");
var import_channel_websocket = __toESM(require("@storybook/channel-websocket"));
var import_core_events = __toESM(require("@storybook/core-events"));
var import_ts_retry = require("ts-retry");

// src/stories.ts
var import_node_path = __toESM(require("path"));
var import_node_fs = require("fs");
var glob = __toESM(require("glob"));
var import_loader = require("@storybook/react-native/scripts/loader.js");
var import_core_common = require("@storybook/core-common");
var csfTools = __toESM(require("@storybook/csf-tools"));
var getCFSStories = ({
  configPath,
  customPaths,
  isAbsolutePaths
}) => {
  const hasCustomPaths = Array.isArray(customPaths);
  if (hasCustomPaths && customPaths.length === 0) {
    die(`\u{1F534} No custom paths provided`);
    process.exit(1);
  }
  let stories;
  if (hasCustomPaths) {
    warn("\u{1F7E0} Using custom paths:");
    customPaths?.forEach((path2) => {
      info(`${path2}`);
    });
    stories = customPaths;
  } else {
    const mainImport = (0, import_loader.getMain)({ configPath });
    const main2 = mainImport.default ?? mainImport;
    stories = main2.stories;
  }
  const storiesSpecifiers = (0, import_core_common.normalizeStories)(stories, {
    configDir: configPath,
    workingDir: process.cwd()
  });
  const storyPaths = storiesSpecifiers.reduce((acc, specifier) => {
    const paths = glob.sync(specifier.files, {
      cwd: import_node_path.default.resolve(process.cwd(), specifier.directory),
      absolute: isAbsolutePaths,
      // default to always ignore (exclude) anything in node_modules
      ignore: ["**/node_modules"]
    }).map((storyPath) => {
      const pathWithDirectory = import_node_path.default.join(specifier.directory, storyPath);
      const requirePath = isAbsolutePaths ? storyPath : ensureRelativePathHasDot(
        import_node_path.default.relative(configPath, pathWithDirectory)
      );
      const normalizePathForWindows = (str) => import_node_path.default.sep === "\\" ? str.replace(/\\/g, "/") : str;
      return normalizePathForWindows(requirePath);
    });
    return [...acc, ...paths];
  }, []);
  if (storyPaths.length === 0) {
    die(`\u{1F534} No stories found`);
    process.exit(1);
  }
  const csfStories = storyPaths.map((storyPath) => {
    const code = (0, import_node_fs.readFileSync)(storyPath, { encoding: "utf-8" }).toString();
    return csfTools.loadCsf(code, {
      fileName: storyPath,
      makeTitle: (userTitle) => userTitle
    }).parse();
  });
  return csfStories;
};

// src/screenshot.ts
var import_looks_same = __toESM(require("looks-same"));
var osnapAndroid = __toESM(require("@ferocia-oss/osnap/src/android.js"));
var osnapIos = __toESM(require("@ferocia-oss/osnap/src/ios.js"));
var takeScreenshot = async ({
  name,
  platform,
  dest
}) => {
  const osnap = platform === "android" ? osnapAndroid : osnapIos;
  info(`\u{1F4F8}${getEmojiForPlatform(platform)} Taking screenshot			${name}`);
  try {
    await osnap.saveToFile({
      platform,
      filename: getScreenshotPath({
        name,
        platform,
        type: dest
      })
    });
  } catch (error2) {
    die(
      `\u{1F534}${getEmojiForPlatform(
        platform
      )} Failed to take screenshot => "${name}"`,
      error2?.message ?? "Unknown error"
    );
    process.exit(1);
  }
};
var compareScreenshots = async ({
  storyId,
  platform
}) => {
  const reference = getScreenshotPath({
    name: storyId,
    platform,
    type: "base"
  });
  const current = getScreenshotPath({
    name: storyId,
    platform,
    type: "current"
  });
  const diff = getScreenshotPath({
    name: storyId,
    platform,
    type: "diff"
  });
  info(
    `\u{1F50D}${getEmojiForPlatform(platform)} Comparing screenshots		${storyId}`
  );
  const { equal } = await (0, import_looks_same.default)(current, reference);
  const isFailed = !equal;
  if (isFailed) {
    await import_looks_same.default.createDiff({
      reference,
      current,
      diff,
      highlightColor: "#ff00ff",
      // color to highlight the differences
      strict: false,
      // strict comparison
      tolerance: 2.5,
      antialiasingTolerance: 0,
      ignoreAntialiasing: true,
      // ignore anti aliasing by default
      ignoreCaret: true
      // ignore caret by default
    });
  }
  return {
    isFailed,
    diff
  };
};

// src/runner.ts
var RunnerEvents = {
  CREATING_CHANNEL: "CREATING_CHANNEL",
  CHANNEL_CREATED: "CHANNEL_CREATED",
  RENDERED_STORY: "RENDERED_STORY",
  TAKING_SHOTS: "TAKING_SHOTS",
  STORY_PROCESSED: "STORY_PROCESSED",
  FINISHED: "FINISHED"
};
var TestRunner = class extends import_node_events.EventEmitter {
  constructor(opts) {
    super();
    this.storyQueue = [];
    this.currentStory = null;
    this.results = [];
    this.settings = {};
    if (opts.mode === COMMANDS.UPDATE) {
      console.error("Update mode is not supported yet!");
      return;
    }
    this.settings = opts;
    this.on(RunnerEvents.RENDERED_STORY, (_storyId) => {
      this.makeScreenshots();
    });
    this.on(RunnerEvents.STORY_PROCESSED, (_storyId) => {
      this.processStory();
    });
    this.emitChange(RunnerEvents.CREATING_CHANNEL);
  }
  start() {
    const channel = (0, import_channel_websocket.default)({
      url: this.settings.url,
      onError: (error2) => {
        if ("message" in error2 && error2.message) {
          die("\u{1F534} WebSocket connection failed", error2.message);
        } else {
          die("\u{1F534} Failed creating error", JSON.stringify(error2));
        }
        process.exit(1);
      }
    });
    import_addons.addons.setChannel(channel);
    channel.emit(import_core_events.default.CHANNEL_CREATED, {
      host: this.settings.host,
      port: this.settings.port,
      secured: this.settings.secured
    });
    const csfStories = getCFSStories({
      configPath: this.settings.configPath,
      isAbsolutePaths: this.settings.absolute,
      customPaths: this.settings.customPaths
    });
    const storyIds = csfStories.map(({ stories }) => stories.map(({ id }) => id)).flat();
    this.storyQueue = [...storyIds];
    channel.on(import_core_events.default.STORY_RENDERED, (id) => {
      if (!this.currentStory) {
        console.error("[channel.on] STORY_RENDERED => No current story!");
        return;
      }
      let story = this.currentStory;
      if (story.id !== id) {
        console.error(
          "[channel.on] STORY_RENDERED => Rendered story is not current story!"
        );
        return;
      }
      story = {
        ...story,
        timeRendered: story.timeRendered + 1
      };
      if (story.timeRendered < this.settings.platforms.length) {
        this.currentStory = { ...story };
      } else {
        this.currentStory = {
          ...story,
          state: "RENDERED"
        };
        this.emitChange(RunnerEvents.RENDERED_STORY, id);
      }
    });
    this.channel = channel;
    this.emitChange(RunnerEvents.CHANNEL_CREATED);
  }
  processStory() {
    const storyId = this.storyQueue.pop();
    if (!storyId) {
      this.emitChange(RunnerEvents.FINISHED, this.results);
      return;
    }
    this.currentStory = {
      id: storyId,
      state: "IDLE",
      timeRendered: 0
    };
    this.channel?.emit(import_core_events.default.FORCE_REMOUNT, { storyId });
    this.channel?.emit(import_core_events.default.SET_CURRENT_STORY, { storyId });
  }
  async makeScreenshots() {
    this.emitChange(RunnerEvents.TAKING_SHOTS);
    const story = this.currentStory;
    if (!story) {
      console.error("[makeScreenshots] No story");
      return;
    }
    for (const platform of this.settings.platforms) {
      const screenshotInfo = {
        name: story.id,
        platform,
        dest: this.settings.mode === COMMANDS.INIT ? "base" : "current"
      };
      await takeScreenshot(screenshotInfo);
      try {
        await (0, import_ts_retry.retryAsync)(
          async () => await isFileExists(
            getScreenshotPath({
              type: screenshotInfo.dest,
              platform,
              name: story.id
            })
          ),
          {
            delay: 200,
            maxTry: 5,
            until: (isExists) => isExists
          }
        );
      } catch (error2) {
        die(`\u{1F534} Screenshot not found => "${story.id}" on ${platform}.`);
        process.exit(1);
      }
      if (this.settings.mode === COMMANDS.TEST) {
        const { isFailed, diff } = await compareScreenshots({
          platform,
          storyId: story.id
        });
        if (isFailed && this.settings.exitOnError) {
          die(
            `\u{1F534} Failed test => "${story.id}" on ${platform}. See diff at ${diff}`
          );
          process.exit(1);
        }
        this.results.push({
          isFailed,
          id: story.id,
          platform,
          diff
        });
      }
    }
    this.currentStory = {
      ...story,
      state: "SHOTS_CAPTURED"
    };
    this.emitChange(RunnerEvents.STORY_PROCESSED, story.id);
  }
  startProcessing() {
    this.processStory();
  }
  emitChange(task, ...args) {
    return this.emit(task, ...args);
  }
  removeListeners() {
    this.removeAllListeners();
    if (this.channel) {
      this.channel.removeAllListeners();
    }
  }
};

// src/main.ts
var main = async () => {
  console.time("Script run time");
  await sayHi();
  const { command, options } = await getCliOptions();
  const { exitOnError, files: customPaths, platforms, force } = options;
  await createDir(TOOL_CONFIG_DIR);
  const doesGitignoreExits = await isFileExists(
    `${TOOL_CONFIG_DIR}/.gitignore`
  );
  if (!doesGitignoreExits) {
    info(`\u{1F7E0} Creating ${TOOL_CONFIG_DIR}/.gitignore`);
    await createFile(`${TOOL_CONFIG_DIR}/.gitignore`, ".diff\n.current\n");
  }
  await createDir(SCREENSHOTS_DIRS.base);
  await createDir(SCREENSHOTS_DIRS.current);
  await createDir(SCREENSHOTS_DIRS.diff);
  const isBaseEmpty = await isDirEmpty(SCREENSHOTS_DIRS.base);
  if (isBaseEmpty && command !== COMMANDS.INIT && command !== COMMANDS.UPDATE) {
    warn(
      `\u{1F7E0} Base screenshots directory is empty. Run "${COMMANDS.INIT}" command first.`
    );
    process.exit(1);
  }
  if (!isBaseEmpty && command === COMMANDS.INIT && !force) {
    warn(
      `\u{1F7E0} Base screenshots directory is not empty. If you want to override the current screenshots, use "${COMMANDS.INIT} --force".`
    );
    process.exit(1);
  }
  const isCurrentEmpty = await isDirEmpty(SCREENSHOTS_DIRS.current);
  if (isCurrentEmpty && command === COMMANDS.UPDATE) {
    warn(
      `\u{1F7E0} No generated screenshots to approve. Run "${COMMANDS.TEST}" command first.`
    );
    process.exit(1);
  }
  await (0, import_fs_extra3.emptyDir)(SCREENSHOTS_DIRS.diff);
  if (command === COMMANDS.UPDATE) {
    warn(`\u{1FA84} Updating base screenshots`);
    await update(platforms);
    return;
  }
  if (command === COMMANDS.INIT) {
    warn(`\u{1FA84} Generating base screenshots`);
  }
  if (command === COMMANDS.TEST) {
    warn(`\u{1F9EA} Testing stories`);
  }
  if (platforms.includes("android")) {
    warn(
      `${getEmojiForPlatform("android")} Reversing adb port ${SERVER_DEFAULTS.port}`
    );
    await reverseAdbPort(SERVER_DEFAULTS.port);
  }
  if (platforms.length === 1 && platforms[0] === "android") {
    await shutdownBootedIosDevices();
  }
  if (platforms.length === 1 && platforms[0] === "ios") {
    await shutdownAndroidEmulator();
  }
  const runner = new TestRunner({
    mode: command,
    platforms,
    url: SERVER_DEFAULTS.url,
    host: SERVER_DEFAULTS.host,
    port: SERVER_DEFAULTS.port,
    secured: SERVER_DEFAULTS.secured,
    absolute: SERVER_DEFAULTS.absolute,
    configPath: STORYBOOK_CONFIG_PATH,
    customPaths,
    exitOnError
  });
  runner.on("CHANNEL_CREATED", () => {
    info(`\u{1F7E2} Storybook channel created`);
    runner.startProcessing();
  });
  runner.on("FINISHED", (results) => {
    if (command === COMMANDS.INIT) {
      success(`\u{1F7E2} Base screenshots generated`);
      info(`See screenshots at ${getScreenshotDir("base")}`);
      process.exit(0);
    }
    results.forEach(({ id, isFailed, platform, diff }) => {
      if (isFailed) {
        die(
          `\u{1F534}${getEmojiForPlatform(platform)} Failed test	${id}`,
          `See diff at ${diff}`
        );
      } else {
        success(`\u{1F7E2}${getEmojiForPlatform(platform)} Passed test	${id}`);
      }
    });
    console.timeEnd("Script run time");
    runner.removeListeners();
    process.exit(results.some(({ isFailed }) => isFailed) ? 1 : 0);
  });
  runner.start();
};

// src/index.ts
main();
